#!/usr/bin/env python

"""
Moin serialiser support.

Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from urllib import quote_plus

class Serialiser:

    "General serialisation support."

    format = None # defined by subclasses

    def __init__(self, output, formats=None, linker=None):

        """
        Initialise the serialiser with an 'output' context, an optional
        'formats' mapping from names to serialiser classes, and an optional
        'linker' object for translating links.
        """

        self.output = output
        self.formats = formats
        self.linker = linker

        # Initialise a callable for use in serialisation.

        self.out = output.out

        # Initialisation of any other state.

        self.init()

    def init(self):

        "Initialisation method to be overridden by subclasses."

        pass

    def __repr__(self):
        return "%s(%r, %r, %r)" % (self.__class__.__name__, self.output,
                                   self.formats, self.linker)

    def get_serialiser(self, format):

        """
        Return a serialiser for the given 'format'. Return self if no suitable
        serialiser can be obtained.
        """

        cls = self.formats and self.formats.get(format)
        if cls:
            return self.instantiate(cls)
        else:
            return self

    def get_output(self):

        "Return the output as a string."

        return self.output.to_string()

    def instantiate(self, cls):

        """
        Instantiate 'cls' and return the result if 'cls' is a different class to
        this instance. Otherwise, return this instance.
        """

        if cls is self.__class__:
            return self
        else:
            return cls(self.output, self.formats, self.linker)

def escape_attr(s):

    "Escape XML document attribute."

    return escape_text(s).replace("'", "&apos;").replace('"', "&quot;")

def escape_text(s):

    "Escape XML document text."

    return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

def make_id(s):

    "Make a suitable identifier for XML element identification."

    # NOTE: This reproduces the Moin algorithm for compatibility.
    # NOTE: There may well be improvements possible, possibly by replacing plus
    # NOTE: with something less cumbersome, even though plus may be unusual in
    # NOTE: things like headings, anyway.

    # The desired output is the following pattern:

    # [A-Za-z][-_:.A-Za-z0-9]*

    # The Python UTF-7 encoder preserves symbols and it encodes + as +- with an
    # output range as follows (in addition to A-Za-z0-9):

    # -_:.%+ !"#$&\'()*,/;<=>?@[]^`{|}

    # The quote_plus function converts space to plus, preserves -_:. and encodes
    # all other symbols (including original occurrences of plus and percent) and
    # non-alphanumeric (ASCII) characters using percent encoding.

    # With colons preserved, the resulting output is in the following range
    # (in addition to A-Za-z0-9):

    # -_:.%+

    # Percent will only occur as an encoding prefix. Plus will only occur as a
    # replacement for space.

    # Combining quote_plus and UTF-7 gives the following range (in addition to
    # A-Za-z0-9):

    # -_:.%+

    # Examples:

    #          UTF-7         quote_plus    replace percent and plus
    # :     -> :          -> :          -> :
    # -     -> -          -> -          -> -
    # .     -> .          -> .          -> .
    # %     -> %          -> %25        -> .25
    # +     -> +-         -> %2B-       -> .2B-
    # _     -> _          -> _          -> _
    # space -> space      -> +          -> _

    # See: RFC2152 - UTF-7 A Mail-Safe Transformation Format of Unicode

    quoted = quote_plus(s.encode("utf-7"), ":").replace("%", ".").replace("+", "_")

    # Ensure that the identifier starts with an alphabetical character.

    if not quoted[0].isalpha():
        return "A%s" % quoted
    else:
        return quoted

# vim: tabstop=4 expandtab shiftwidth=4
